@@ -24,8 +24,9 @@ import logging
 import os
 import glob
 import multiprocessing
-
 import cdist
+from cdist.mputil import mp_pool_run
+from . import util
 
 '''
 common:
@@ -78,6 +79,10 @@ class Explorer(object):
             '__target_hostname': self.target_host[1],
             '__target_fqdn': self.target_host[2],
             '__explorer': self.remote.global_explorer_path,
+            '__target_host_tags': self.local.target_host_tags,
+            '__cdist_log_level': util.log_level_env_var_val(self.log),
+            '__cdist_log_level_name': util.log_level_name_env_var_val(
+                self.log),
         }
         self._type_explorers_transferred = []
         self.jobs = jobs
@@ -96,7 +101,7 @@ class Explorer(object):
         out_path directory.
 
         """
-        self.log.info("Running global explorers")
+        self.log.verbose("Running global explorers")
         self.transfer_global_explorers()
         if self.jobs is None:
             self._run_global_explorers_seq(out_path)
@@ -110,29 +115,23 @@ class Explorer(object):
             fd.write(output)
 
     def _run_global_explorers_seq(self, out_path):
-        self.log.info("Running global explorers sequentially")
+        self.log.debug("Running global explorers sequentially")
         for explorer in self.list_global_explorer_names():
             self._run_global_explorer(explorer, out_path)
 
     def _run_global_explorers_parallel(self, out_path):
-        self.log.info("Running global explorers in {} parallel jobs".format(
+        self.log.debug("Running global explorers in {} parallel jobs".format(
             self.jobs))
-        self.log.debug("Multiprocessing start method is {}".format(
+        self.log.trace("Multiprocessing start method is {}".format(
             multiprocessing.get_start_method()))
-        self.log.debug(("Starting multiprocessing Pool for global "
+        self.log.trace(("Starting multiprocessing Pool for global "
                        "explorers run"))
-        with multiprocessing.Pool(self.jobs) as pool:
-            self.log.debug("Starting async for global explorer run")
-            results = [
-                pool.apply_async(self._run_global_explorer, (e, out_path,))
-                for e in self.list_global_explorer_names()
-            ]
-
-            self.log.debug("Waiting async results for global explorer runs")
-            for r in results:
-                r.get()  # self._run_global_explorer returns None
-            self.log.debug(("Multiprocessing run for global explorers "
-                           "finished"))
+        args = [
+            (e, out_path, ) for e in self.list_global_explorer_names()
+        ]
+        mp_pool_run(self._run_global_explorer, args, jobs=self.jobs)
+        self.log.trace(("Multiprocessing run for global explorers "
+                        "finished"))
 
     # logger is not pickable, so remove it when we pickle
     def __getstate__(self):
@@ -170,20 +169,27 @@ class Explorer(object):
         except EnvironmentError:
             return []
 
-    def run_type_explorers(self, cdist_object):
+    def run_type_explorers(self, cdist_object, transfer_type_explorers=True):
         """Run the type explorers for the given object and save their output
         in the object.
 
         """
-        self.log.debug("Transfering type explorers for type: %s",
-                       cdist_object.cdist_type)
-        self.transfer_type_explorers(cdist_object.cdist_type)
-        self.log.debug("Transfering object parameters for object: %s",
+        self.log.verbose("Running type explorers for {}".format(
+            cdist_object.cdist_type))
+        if transfer_type_explorers:
+            self.log.trace("Transfering type explorers for type: %s",
+                           cdist_object.cdist_type)
+            self.transfer_type_explorers(cdist_object.cdist_type)
+        else:
+            self.log.trace(("No need for transfering type explorers for "
+                            "type: %s"),
+                           cdist_object.cdist_type)
+        self.log.trace("Transfering object parameters for object: %s",
                        cdist_object.name)
         self.transfer_object_parameters(cdist_object)
         for explorer in self.list_type_explorer_names(cdist_object.cdist_type):
             output = self.run_type_explorer(explorer, cdist_object)
-            self.log.debug("Running type explorer '%s' for object '%s'",
+            self.log.trace("Running type explorer '%s' for object '%s'",
                            explorer, cdist_object.name)
             cdist_object.explorers[explorer] = output
 
@@ -208,19 +214,22 @@ class Explorer(object):
     def transfer_type_explorers(self, cdist_type):
         """Transfer the type explorers for the given type to the
            remote side."""
-        if cdist_type.explorers:
-            if cdist_type.name in self._type_explorers_transferred:
-                self.log.debug("Skipping retransfer of type explorers for: %s",
-                               cdist_type)
-            else:
-                source = os.path.join(self.local.type_path,
-                                      cdist_type.explorer_path)
-                destination = os.path.join(self.remote.type_path,
-                                           cdist_type.explorer_path)
-                self.remote.mkdir(destination)
-                self.remote.transfer(source, destination)
-                self.remote.run(["chmod", "0700", "%s/*" % (destination)])
-                self._type_explorers_transferred.append(cdist_type.name)
+        try:
+            if cdist_type.explorers:
+                if cdist_type.name in self._type_explorers_transferred:
+                    self.log.trace(("Skipping retransfer of type explorers "
+                                    "for: %s"), cdist_type)
+                else:
+                    source = os.path.join(self.local.type_path,
+                                          cdist_type.explorer_path)
+                    destination = os.path.join(self.remote.type_path,
+                                               cdist_type.explorer_path)
+                    self.remote.mkdir(destination)
+                    self.remote.transfer(source, destination)
+                    self.remote.run(["chmod", "0700", "%s/*" % (destination)])
+                    self._type_explorers_transferred.append(cdist_type.name)
+        except cdist.Error as e:
+            raise cdist.CdistObjectError(cdist_object, e)
 
     def transfer_object_parameters(self, cdist_object):
         """Transfer the parameters for the given object to the remote side."""
